/*
  Copyright (c) 2015, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "generator.hpp"

#include <queue>


namespace solar {

std::set<item_t> generator_t::get_set(const std::string &nonterm) {
	std::set<item_t> set;

	auto entries = nonterms.equal_range(nonterm);
	for (auto entry = entries.first; entry != entries.second; ++entry)
		set.insert(rules[entry->second]);

	return set;
}

void generator_t::close_set(std::set<item_t> *set) {
	std::queue<item_t> queue;
	for (const item_t &item : *set)
		queue.emplace(item);

	for (; !queue.empty(); queue.pop()) {
		const item_t &cur = queue.front();
		if (!cur.has_next())
			continue;

		const symbol_t &sym = cur.get_next_symbol();
		if (sym.get_type() != SYMBOL_TYPE_NONTERM)
			continue;

		for (const item_t &item : get_set(sym.get_value())) {
			if (set->insert(item).second)
				queue.push(item);
		}
	}
}

void generator_t::generate_itemsets() {
	std::queue<std::pair<std::set<item_t>, size_t>> queue;

	{
		std::set<item_t> set0 = get_set("");
		close_set(&set0);

		queue.push(*add_set(set0).first);
	}

	for (; !queue.empty(); queue.pop()) {
		const auto &cur = queue.front();

		const std::set<item_t> empty_set;
		std::map<symbol_t, std::set<item_t>> new_sets;

		for (const item_t &item : cur.first) {
			if (!item.has_next())
				continue;

			const symbol_t &sym = item.get_next_symbol();

			item_t shifted = item;
			shifted.shift();

			std::set<item_t> &set = new_sets.emplace(sym, empty_set).first->second;
			set.insert(std::move(shifted));
		}

		for (auto &entry : new_sets) {
			close_set(&entry.second);

			auto added = add_set(entry.second);

			if (entry.first.get_type() == SYMBOL_TYPE_NONTERM)
				gotos.emplace(std::make_pair(cur.second, entry.first), added.first->second);
			else
				actions.emplace(std::make_pair(cur.second, entry.first), std::make_pair('s', added.first->second));

			if (added.second)
				queue.push(*added.first);
		}

		for (const item_t &item : cur.first) {
			auto it = rule_ids.find(item);
			if (it != rule_ids.end()) {
				//if (it->second) {
				//	for (const symbol_t &term : terminals)
				//		transitions.emplace(std::make_pair(cur.second, term), std::make_pair('r', it->second));
				//}

				actions.emplace(std::make_pair(cur.second, symbol_t::make_term("")), std::make_pair('r', it->second));
			}
		}
	}
}

void generator_t::print_symbol(const symbol_t &sym) {
	switch (sym.get_type()) {
	case SYMBOL_TYPE_CHAR:
		printf("'%c'", sym.get_value()[0]);
		break;

	default:
		printf("%s", sym.get_value().c_str());
	}
}

void generator_t::print_item(const item_t &item, bool point) {
	printf("%s |=", item.get_lhs().c_str());

	for (size_t i = 0; i <= item.get_rhs().size(); i++) {
		if (i == item.get_point() && point)
			printf(" .");

		if (i == item.get_rhs().size())
			break;

		printf(" ");
		print_symbol(item.get_rhs()[i]);
	}

	printf("\n");
}

void generator_t::print_set(const std::set<item_t> &set) {
	for (const item_t &item : set)
		print_item(item, true);
}

generator_t::generator_t(const std::vector<item_t> &rules0) : rules(rules0) {
	for (size_t i = 0; i < rules.size(); i++) {
		item_t rule = rules[i];

		nonterms.emplace(rule.get_lhs(), i);

		while (rule.has_next()) {
			const symbol_t &sym = rule.get_next_symbol();
			items.emplace(sym, rule);

			if (sym.get_type() != SYMBOL_TYPE_NONTERM)
				terminals.insert(sym);

			rule.shift();
		}

		rule_ids.emplace(rule, i);
	}

	generate_itemsets();

	printf("Rules:\n");
	for (size_t i = 0; i < rules.size(); i++) {
		printf("(%u) ", unsigned(i));
		print_item(rules[i], false);
	}
	printf("\n");

	const std::set<item_t> *setlist[itemsets.size()];
	for (const auto &entry : itemsets)
		setlist[entry.second] = &entry.first;

	for (size_t i = 0; i < itemsets.size(); i++) {
		printf("Item %u:\n", unsigned(i));
		print_set(*setlist[i]);
		printf("\n");
	}

	printf("Actions:\n");
	for (const auto &a : actions) {
		printf("%u ", unsigned(a.first.first));

		if (a.second.second) {
			print_symbol(a.first.second);
			printf(" -> %c%u\n", a.second.first, unsigned(a.second.second));
		}
		else {
			printf("$ -> acc\n");
		}
	}
	printf("\n");

	printf("Gotos:\n");
	for (const auto &g : gotos)
		printf("%u %s -> %u\n", unsigned(g.first.first), g.first.second.get_value().c_str(), unsigned(g.second));
}

}
